""" A kernel transformation that precomputes data and then splits computation
in chunks of vector size independent of divisibility of the loop bounds. """

from dune.perftool.loopy.vcl import get_vcl_type_size
from dune.perftool.loopy.transformations.vectorview import (add_vector_view,
                                                            get_vector_view_name,
                                                            )
from dune.perftool.tools import get_pymbolic_basename

from loopy.kernel.creation import parse_domains
from loopy.symbolic import pw_aff_to_expr

from pymbolic.mapper.dependency import DependencyMapper
from pymbolic.mapper.substitutor import substitute

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def collect_vector_data_precompute(knl, insns, inames):
    #
    # Process/Assert/Standardize the input
    #

    # inames input -> tuple
    if isinstance(inames, str):
        inames = inames.split(",")
    inames = tuple(i.strip() for i in inames)

    # insns -> list of Instruction instances
    if isinstance(insns, lp.match.MatchExpressionBase):
        insns = lp.find_instructions(knl, insns)
    else:
        if isinstance(insns, str):
            insns = [i.strip() for i in insns.split(",")]
        insns = [knl.id_to_insn[i] for i in insns]

    #
    # Inspect the given instructions for dependent quantities
    # and precompute them
    #

    quantities = []
    for insn in insns:
        for expr in DependencyMapper()(insn.expression):
            quantities.append(get_pymbolic_basename(expr))
    quantities = set(quantities)

    for quantity in quantities:
        # Introduce a substitution rule and find save name
        subst_old = knl.substitutions.keys()
        knl = lp.assignment_to_subst(knl, quantity, extra_arguments=inames)
        subst_new = knl.substitutions.keys()
        subst_name, = set(subst_new) - set(subst_old)

        # Do precomputation of the quantity
        prec_quantity = "{}_precomputed".format(quantity)
        knl = lp.precompute(knl, subst_name, inames,
                            temporary_name=prec_quantity,
                            )

        # Introduce a vector view of the precomputation result
        knl = add_vector_view(knl, prec_quantity)

    #
    # Construct a flat loop for the given instructions
    #

    new_insns = []
    other_insns = [i for i in knl.instructions if i.id not in [j.id for j in insns]]

    size = prim.Product(tuple(pw_aff_to_expr(knl.get_iname_bounds(i).size) for i in inames))
    vec_size = get_vcl_type_size(np.float64)
    size = prim.FloorDiv(size, vec_size)

    temporaries = knl.temporary_variables
    temporaries["flatsize"] = lp.TemporaryVariable("flatsize",
                                                   dtype=np.int32,
                                                   shape=(),
                                                   )
    new_insns.append(lp.Assignment(prim.Variable("flatsize"),
                                   size,
                                   )
                     )

    # Add an additional domain to the kernel
    new_iname = "flat_{}".format("_".join(inames))
    domain = "{{ [{0}] : 0<={0}<flatsize }}".format(new_iname, str(size))
    domain = parse_domains(domain, {})
    knl = knl.copy(domains=knl.domains + domain,
                   temporary_variables=temporaries)

    # Split and tag the flat iname
    knl = lp.split_iname(knl, new_iname, vec_size, inner_tag="vec")
    new_inames = ("{}_outer".format(new_iname), "{}_inner".format(new_iname))
    knl = lp.assume(knl, "flatsize mod {} = 0".format(vec_size))

    for insn in insns:
        # Get a vector view of the lhs expression
        lhsname = get_pymbolic_basename(insn.assignee)
        knl = add_vector_view(knl, lhsname)
        lhsname = get_vector_view_name(lhsname)

        new_insns.append(lp.Assignment(prim.Subscript(prim.Variable(lhsname), tuple(prim.Variable(i) for i in new_inames)),
                                       prim.Subscript(prim.Variable(get_vector_view_name("wk_precomputed")), tuple(prim.Variable(i) for i in new_inames)),
                                       within_inames=frozenset(new_inames),
                                       within_inames_is_final=True,
                                       )
                         )

    return knl.copy(instructions=new_insns + other_insns)
